"""
Tools for working with code.
"""

import sys
import os
import io
import inspect
import ast
import textwrap
import importlib
from types import LambdaType
from collections import defaultdict, OrderedDict
from tokenize import NAME, tokenize, untokenize

import networkx as nx


def _get_long_name(node):
    # If the node is an Attribute or Name node that is composed
    # only of other Attribute or Name nodes, then return the full
    # dotted name for this node. Otherwise, i.e., if this node
    # contains Subscripts or Calls, return None.
    if isinstance(node, ast.Name):
        return node.id
    elif not isinstance(node, ast.Attribute):
        return None
    val = node.value
    parts = [node.attr]
    while True:
        if isinstance(val, ast.Attribute):
            parts.append(val.attr)
            val = val.value
        elif isinstance(val, ast.Name):
            parts.append(val.id)
            break
        else:  # it's more than just a simple dotted name
            return None
    return '.'.join(parts[::-1])


class _SelfCallCollector(ast.NodeVisitor):
    """
    An ast.NodeVisitor that records calls to self.* methods.
    """

    def __init__(self, class_):
        super().__init__()
        self.self_calls = defaultdict(list)
        self.class_ = class_
        self.mro = inspect.getmro(class_)
        self.mro_names = set([c.__name__ for c in self.mro])

    def visit_Call(self, node):  # (func, args, keywords, starargs, kwargs)
        fncname = _get_long_name(node.func)
        class_ = self.class_
        if fncname is not None:
            if fncname.startswith('self.') and len(fncname.split('.')) == 2:
                shortfnc = fncname.split('.')[1]
                if shortfnc not in self.self_calls[class_]:
                    self.self_calls[class_].append(shortfnc)
                for arg in node.args:
                    self.visit(arg)
            # check for Class.func(inst) form for base class method call
            elif (len(fncname.split('.')) == 2 and fncname.split('.')[0] in self.mro_names and
                  node.args and isinstance(node.args[0], ast.Name) and node.args[0].id == 'self'):
                cname, func = fncname.split('.')
                for c in self.mro:
                    if c.__name__ == cname:
                        sub_mro = inspect.getmro(c)
                        for sub_c in sub_mro:
                            if func in sub_c.__dict__:
                                c = sub_c
                                break
                        if func not in self.self_calls[c]:
                            self.self_calls[c].append(func)
                        for arg in node.args:
                            self.visit(arg)
                        break
                else:
                    self.generic_visit(node)
            else:
                self.generic_visit(node)
        # check for super() call
        elif isinstance(node.func, ast.Attribute) and isinstance(node.func.value, ast.Call):
            callnode = node.func.value
            n = _get_long_name(callnode.func)
            # if this is a 'super' call, get the base of the specified class
            if n == 'super':  # this only works for a single call level
                if len(callnode.args) == 0:
                    sup_0 = self.mro[0].__name__
                    visit_super = True
                else:
                    sup_1 = _get_long_name(callnode.args[1])
                    sup_0 = _get_long_name(callnode.args[0])
                    visit_super = (sup_1 == 'self' and
                                   sup_0 is not None and len(sup_0.split('.')) == 1)

                if visit_super:
                    for i, c in enumerate(self.mro[:-1]):
                        if sup_0 == c.__name__:
                            # we need super of the specified class
                            sub_mro = inspect.getmro(c)
                            for sub_c in sub_mro:
                                if sub_c is not c:
                                    c = sub_c
                                    break
                            fn = node.func.attr
                            if fn not in self.self_calls[c]:
                                self.self_calls[c].append(fn)
                            for arg in node.args:
                                self.visit(arg)
                            break
                    else:
                        self.generic_visit(node)
            else:
                self.generic_visit(node)
        else:
            self.generic_visit(node)


def _find_owning_class(mro, func_name):
    """
    Return the full funcname and class where the function is first found in the class MRO.
    """
    # TODO: this won't work for classes with __slots__

    for c in mro:
        if func_name in c.__dict__:
            return '.'.join((c.__name__, func_name)), c

    return None, None


def _get_nested_calls(starting_class, class_, func_name, parent, graph, seen):
    """
    Parse the AST of the given method and all 'self' methods it calls and record owning classes.
    """
    func = getattr(class_, func_name)
    src = inspect.getsource(func)
    dedented_src = textwrap.dedent(src)

    node = ast.parse(dedented_src, mode='exec')
    visitor = _SelfCallCollector(starting_class)
    visitor.visit(node)

    seen.add('.'.join((class_.__name__, func_name)))

    # now find the actual owning class for each call
    for klass, funcset in visitor.self_calls.items():
        mro = inspect.getmro(klass)
        for f in funcset:
            full, c = _find_owning_class(mro, f)
            if full is not None:
                graph.add_edge(parent, full)
                if full not in seen:
                    _get_nested_calls(starting_class, c, f, full, graph, seen)


def get_nested_calls(class_, method_name, stream=sys.stdout):
    """
    Display the call tree for the specified class method and all 'self' class methods it calls.

    Parameters
    ----------
    class_ : class
        The starting class.
    method_name : str
        The name of the class method.
    stream : file-like
        The output stream where output will be displayed.

    Returns
    -------
    networkx.DiGraph
        A graph containing edges from methods to their sub-methods.
    """
    # moved this class def in here to keep the numpy doc scraper from barfing due to
    # stuff in nx.DiGraph.
    class OrderedDiGraph(nx.DiGraph):
        """
        A DiGraph using OrderedDicts for internal storage.
        """

        node_dict_factory = OrderedDict
        adjlist_dict_factory = OrderedDict
        edge_attr_dict_factory = OrderedDict

    graph = OrderedDiGraph()
    seen = set()
    top = object()

    full, klass = _find_owning_class(inspect.getmro(class_), method_name)
    if full is None:
        print("Can't find function '%s' in class '%s'." % (method_name, class_.__name__))
    else:
        graph.add_edge(top, full)
        parent = full
        _get_nested_calls(class_, klass, method_name, parent, graph, seen)

    if graph and stream is not None:
        seen = set([top])
        stack = [(0, iter(graph[top]))]
        while stack:
            depth, children = stack[-1]
            try:
                n = next(children)
                stream.write("%s%s\n" % ('  ' * depth, n))
                if n not in seen:
                    stack.append((depth + 1, iter(graph[n])))
                    seen.add(n)
            except StopIteration:
                stack.pop()

    return graph


def _calltree_setup_parser(parser):
    """
    Set up the command line options for the 'openmdao call_tree' command line tool.
    """
    parser.add_argument('method_path', nargs=1,
                        help='Full module path to desired class method, e.g., '
                        '"openmdao.components.exec_comp.ExecComp.setup".')
    parser.add_argument('-o', '--outfile', action='store', dest='outfile',
                        default='stdout', help='Output file.  Defaults to stdout.')


def _calltree_exec(options, user_args):
    """
    Process command line args and call get_nested_calls on the specified class method.
    """
    parts = options.method_path[0].split('.')
    if len(parts) < 3:
        raise RuntimeError("You must supply the full module path to the function, "
                           "for example:  openmdao.api.Group._setup.")

    class_name = parts[-2]
    func_name = parts[-1]
    modpath = '.'.join(parts[:-2])

    old_syspath = sys.path[:]
    sys.path.append(os.getcwd())

    try:
        mod = importlib.import_module(modpath)
        klass = getattr(mod, class_name)

        stream_map = {'stdout': sys.stdout, 'stderr': sys.stderr}
        stream = stream_map.get(options.outfile)
        if stream is None:
            stream = open(options.outfile, 'w')

        get_nested_calls(klass, func_name, stream)
    finally:
        sys.path = old_syspath


def _target_iter(targets):
    for target in targets:
        if isinstance(target, ast.Tuple):
            for t in target.elts:
                yield t
        else:
            yield target


class _AttrCollector(ast.NodeVisitor):
    """
    An ast.NodeVisitor that records class attribute names.
    """

    def __init__(self, class_dict):
        super().__init__()
        self.class_dict = class_dict
        self.class_stack = []
        self.func_stack = []
        self.names = None
        self.decnames = None

    def get_attributes(self):
        return self.class_dict

    def visit_ClassDef(self, node):
        full_name = '.'.join(self.class_stack[:] + [node.name])
        self.class_stack.append(full_name)
        self.class_dict[full_name] = set()
        for stmt in node.body:
            self.visit(stmt)
        self.class_stack.pop()

        if self.func_stack:  # ignore classes nested in functs
            del self.class_dict[full_name]

    def visit_FunctionDef(self, node):
        self.func_stack.append(node.name)
        for stmt in node.body:
            self.visit(stmt)
        self.func_stack.pop()

        if self.class_stack:
            # see if this is a property, and if so, treat as an attribute
            for dec in node.decorator_list:
                self.decnames = []
                self.visit(dec)
                if len(self.decnames) == 1 and self.decnames[0] == 'property':
                    self.class_dict[self.class_stack[-1]].add(node.name)

        self.decnames = None

    def visit_Assign(self, node):
        if self.class_stack:
            for t in _target_iter(node.targets):
                self.names = []
                self.visit(t)
                if len(self.names) > 1 and self.names[0] == 'self':
                    self.class_dict[self.class_stack[-1]].add(self.names[1])

            self.names = None

    def visit_Attribute(self, node):
        if self.names is not None:
            self.visit(node.value)
            self.names.append(node.attr)

    def visit_Name(self, node):
        if self.names is not None:
            self.names.append(node.id)
        elif self.decnames is not None:
            self.decnames.append(node.id)


def get_class_attributes(fname, class_dict=None):
    """
    Find all referenced attributes in all classes defined in the given file.

    Parameters
    ----------
    fname : str
        File name.
    class_dict : dict or None
        Dict mapping class names to attribute names.

    Returns
    -------
    dict
        The dict maps class name to a set of attribute names.
    """
    if class_dict is None:
        class_dict = {}

    with open(fname, 'r') as f:
        source = f.read()
        node = ast.parse(source, mode='exec')
        visitor = _AttrCollector(class_dict)
        visitor.visit(node)
        return visitor.get_attributes()


def _get_return_name(node):
    return node.id if isinstance(node, ast.Name) else None


def _get_return_names(outs):
    """
    Return a list of (name or None) for each return value.

    If there are multiple returns that differ by name or number of return values, an exception
    will be raised.  If one entry in one return list has a name and another is None, the name
    will take precedence and no exception will be raised.

    Returns
    -------
    list
        The list of return names.  Some entries will be None if there was no simple name
        associated with a given return value.
    """
    if len(outs) == 0:
        return []
    if len(outs) == 1:
        return outs[0]

    names = outs[0].copy()
    length = len(names)
    for lst in outs[1:]:
        if len(lst) != length:
            raise RuntimeError("Function has multiple return statements with differing numbers "
                               "of return values.")

        for i, (name, newname) in enumerate(zip(names, lst)):
            if name is None:
                names[i] = newname
            elif newname is not None and name != newname:
                raise RuntimeError("Function has multiple return statements with different "
                                   f"return value names of {sorted((name, newname))} for "
                                   f"return value {i}.")
    return names


def get_return_names(func):
    """
    Return the names of the variables returned by the given function.

    Returns None for any return values that aren't a simple name.

    Parameters
    ----------
    func : function
        The function to be examined.

    Returns
    -------
    list
        The names of the variables returned by the given function.
    """
    class _FuncRetNameCollector(ast.NodeVisitor):
        """
        An ast.NodeVisitor that records return value names.

        Attributes
        ----------
        _ret_infos : list
            List containing one entry for each return statement, with each entry containing a list
            of name (or None) for each function return value.
        """

        def __init__(self, func):
            super().__init__()
            self._ret_infos = []
            self.visit(ast.parse(textwrap.dedent(inspect.getsource(func)), mode='exec'))

        def get_return_names(self):
            """
            Return a list of (name or None) for each return value.

            If there are multiple returns that differ by name or number of return values, an
            exception will be raised.  If one entry in one return list has a name and another is
            None, the name will take precedence and no exception will be raised.

            Returns
            -------
            list
                The list of return names.  Some entries will be None if there was no simple name
                associated with a given return value.
            """
            return _get_return_names(self._ret_infos)

        def visit_Return(self, node):
            """
            Visit a Return node.

            Parameters
            ----------
            node : ASTnode
                The return node being visited.
            """
            self._ret_infos.append([])

            if isinstance(node.value, ast.Tuple):
                for n in node.value.elts:
                    self._ret_infos[-1].append(_get_return_name(n))
            else:
                self._ret_infos[-1].append(_get_return_name(node.value))

    return _FuncRetNameCollector(func).get_return_names()


class _FuncGrapher(ast.NodeVisitor):
    """
    An ast.NodeVisitor that builds a graph between a function's inputs and outputs.
    """

    def __init__(self, node):
        super().__init__()
        self.rhs = []
        self.lhs = []
        self.names = None
        self.graph = nx.DiGraph()
        self.outs = []
        self.fstack = []
        self.visit(node)

    def _update_graph(self):
        for inp in self.rhs:
            for out in self.lhs:
                self.graph.add_edge(inp, out)
        self.lhs = []
        self.rhs = []

    def visit_FunctionDef(self, node):
        if self.fstack:
            raise RuntimeError("Function contains nested functions, which are not supported.")
        self.fstack.append(node)
        for stmt in node.body:
            self.visit(stmt)
        self.fstack.pop()

    def visit_Assign(self, node):
        self.names = self.lhs
        for t in _target_iter(node.targets):
            self.visit(t)
        self.names = self.rhs
        self.visit(node.value)
        self.names = None
        self._update_graph()

    def visit_Attribute(self, node):
        pass  # skip any Name nodes that are part of an Attribute node

    def visit_Call(self, node):
        for arg in node.args:
            self.visit(arg)

    def visit_Name(self, node):
        if self.names is not None:
            self.names.append(node.id)

    def visit_Return(self, node):
        self.outs.append([])

        if isinstance(node.value, ast.Tuple):
            it = enumerate(node.value.elts)
        else:
            it = [(0, node.value)]

        for i, n in it:
            self.lhs = [f"@out{i}"]
            self.rhs = []
            self.names = self.rhs
            self.visit(n)
            self._update_graph()
            self.outs[-1].append(_get_return_name(n))


def get_func_graph(func, outnames=None, display=False):
    """
    Generate a graph between a function's inputs and outputs.

    Uses the AST to analyze the function and build a graph between inputs and outputs, so the
    function source must be available.

    Parameters
    ----------
    func : Callable
        The function to be analyzed.
    outnames : list or None
        The list of expected output variable names.
    display : bool
        If True, display the graph using pydot.

    Returns
    -------
    networkx.DiGraph
        A graph containing edges from inputs to outputs.
    """
    node = ast.parse(textwrap.dedent(inspect.getsource(func)), mode='exec')
    visitor = _FuncGrapher(node)
    retnames = _get_return_names(visitor.outs)
    inputs = set(inspect.signature(func).parameters)

    # check vs outnames
    if outnames is not None:
        if len(retnames) != len(outnames):
            raise RuntimeError("Number of return values in function does not match number of "
                               f"expected return names. ({outnames}) != ({retnames})")
        for ret, name in zip(retnames, outnames):
            if ret is not None and ret != name:
                raise RuntimeError(f"Return value name '{name}' in function does not match "
                                   f"expected name '{ret}.")
    else:
        outnames = []
        for i, ret in enumerate(retnames):
            if ret is None or ret in inputs:
                outnames.append(f'out{len(outnames)}')
            else:
                outnames.append(ret)

    mapping = {f'@out{i}': name for i, name in enumerate(outnames)}
    visitor.graph = nx.relabel_nodes(visitor.graph, mapping)
    to_remove = [e for e in visitor.graph.edges() if e[0] == e[1]]
    visitor.graph.remove_edges_from(to_remove)
    # make sure all outputs exist as nodes in graph, even if they have no incoming edges
    visitor.graph.add_nodes_from(outnames)
    visitor.graph.graph['inputs'] = inputs
    visitor.graph.graph['outputs'] = outnames

    if display:
        # show the function graph visually
        from openmdao.visualization.graph_viewer import write_graph, _to_pydot_graph
        write_graph(_to_pydot_graph(visitor.graph))

    return visitor.graph


def get_partials_deps(func, outputs=None):
    """
    Generate tuples of the form (output, input) for the given function.

    Only tuples where the output depends on the input are yielded. This can be used to
    determine which partials need to be declared.

    Note that currently the function grapher doesn't recurse into functions and assumes that all
    outputs of a sub-function are dependent on all inputs to that function. This may lead to
    a conservative estimate of partials that need to be declared.

    Parameters
    ----------
    func : Callable
        The function to be analyzed.
    outputs : list or None
        The list of output variable names.

    Yields
    ------
    tuple
        A tuple of the form (output, input).
    """
    graph = get_func_graph(func, outputs)
    outs = graph.graph['outputs']
    successors = graph.successors

    for start in graph.graph['inputs']:
        visited = set([start])
        stack = [(start, successors(start))]
        while stack:
            _, succs = stack[-1]
            for succ in succs:
                if succ not in visited:
                    visited.add(succ)
                    if succ in outs:
                        yield succ, start
                    stack.append((succ, successors(succ)))
                    break
            else:
                stack.pop()


def block_filter(tokiter, blocks_to_remove, block_start_tok):
    """
    Remove blocks of code from a stream of tokens.

    Blocks are removed based on indentation level.  If a block's name matches one in
    blocks_to_remove, all non-blank lines where the first token is indented to a greater level
    than the block start token are removed.

    Parameters
    ----------
    tokiter : iterator
        Iterator of tokens.
    blocks_to_remove : set
        Set of block names to remove.
    block_start_tok : str
        The name of the block start token, e.g., 'def' or 'class'.

    Yields
    ------
    tuple
        The next token in the stream, unless it is part of a block that
        should be removed.
    """
    indent = None
    save = []
    for tok in tokiter:
        toktype, tokval, start, _, _ = tok
        tokcol = start[1]
        if save:  # we're on block start line after block start token
            if toktype == NAME and tokval not in blocks_to_remove:
                indent = None
                yield from save
                yield tok
            save = []
            continue
        elif toktype == NAME and tokval == block_start_tok:  # block start line
            indent = tokcol
            save.append(tok)  # we might need to emit this token if block doesn't match
            continue
        elif indent is not None:
            if tokcol > indent or not tokval.strip():  # skip lines that are indented or blank
                continue
            else:  # block is done
                indent = None

        yield tok


def find_block_start(srccode, block_name, block_start_tok):
    """
    Find the start of a block of code.

    Parameters
    ----------
    srccode : str
        Source code to search for block.
    block_name : str
        The name of the block to find.
    block_start_tok : str
        The name of the block start token, e.g., 'def' or 'class'.

    Returns
    -------
    tuple
        A tuple of the form (line number, column number, block start line) or (None, None, None) if
        the block was not found.
    """
    tokiter = tokenize(io.BytesIO(srccode.encode('utf-8')).readline)
    for tok in tokiter:
        toktype, tokval, start, _, _ = tok

        if toktype == NAME and tokval == block_start_tok:  # block start line
            try:
                nxt = next(tokiter)
            except StopIteration:
                return (None, None, None)
            ntoktype, ntokval, _, _, _ = nxt
            if ntoktype == NAME and ntokval == block_name:
                return start[0], start[1], tok[-1]

    return (None, None, None)


def remove_src_blocks(srccode, names, block_start_tok):
    """
    Remove blocks from a piece of source code.

    Parameters
    ----------
    srccode : str
        The source code.
    names : list of str
        List of blocks to be removed.
    block_start_tok : str
        The name of the block start token, e.g., 'def' or 'class'.

    Returns
    -------
    str
        The modified source code.
    """
    if not names:
        return srccode
    return untokenize(block_filter(tokenize(io.BytesIO(srccode.encode('utf-8')).readline),
                                   set(names), block_start_tok=block_start_tok)).decode()


def replace_src_block(srccode, block_name, new_block, block_start_tok):
    """
    Replace a block in a piece of source code.

    Parameters
    ----------
    srccode : str
        The source code.
    block_name : str
        The name of the block to be replaced.
    new_block : str
        The replacement block.
    block_start_tok : str
        The name of the block start token, e.g., 'def' or 'class'.

    Returns
    -------
    str
        The modified source code.
    """
    linenum, _, _ = find_block_start(srccode, block_name, block_start_tok)
    if linenum is None:
        raise RuntimeError(f"Block '{block_start_tok} {block_name}' not found in source code.")

    stream = io.StringIO(srccode)
    lines = srccode.splitlines()
    linenum -= 1  # i is zero indexed, so adjust linenum to be zero indexed
    for i, line in enumerate(lines):
        if i == linenum:
            # insert new block
            for subline in new_block.splitlines():
                print(subline, file=stream)
            print('', file=stream)
        print(line, file=stream)

    newsrc = stream.getvalue()
    # now remove the old block
    return untokenize(block_filter(tokenize(io.BytesIO(newsrc.encode('utf-8')).readline),
                                   {block_name}, block_start_tok=block_start_tok)).decode()


def is_lambda(f):
    """
    Return True if the given function is a lambda function.

    Parameters
    ----------
    f : function
        The function to check.

    Returns
    -------
    bool
        True if the given function is a lambda function.
    """
    return isinstance(f, LambdaType) and f.__name__ == "<lambda>"


class LambdaPickleWrapper(object):
    """
    A wrapper for a lambda function that allows it to be pickled.

    Parameters
    ----------
    lambda_func : function
        The lambda function to be wrapped.

    Attributes
    ----------
    _func : function
        The lambda function.
    _src : str
        The isolated source of the lambda function.
    """

    def __init__(self, lambda_func):
        """
        Initialize the wrapper.

        Parameters
        ----------
        lambda_func : function
            The lambda function to be wrapped.
        """
        self._func = lambda_func
        self._src = None

    def __call__(self, *args, **kwargs):
        """
        Call the lambda function.

        Parameters
        ----------
        *args : list
            Positional arguments.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        object
            The result of the lambda function.
        """
        return self._func(*args, **kwargs)

    def __getstate__(self):
        """
        Return the state of this object for pickling.

        The lambda function is converted to a string for pickling.

        Returns
        -------
        dict
            The state of this object.
        """
        state = self.__dict__.copy()
        state['_func'] = self._getsrc()
        return state

    def __setstate__(self, state):
        """
        Restore the state of this object after pickling.

        Parameters
        ----------
        state : dict
            The state of this object.
        """
        self.__dict__.update(state)
        self._func = eval(state['_func'])  # nosec

    def _getsrc(self):
        if self._src is None:
            self._src = _LambdaSrcFinder(self._func).src
            if self._src is None:
                raise RuntimeError("The fix for pickling lambda functions only works for python "
                                   "3.9 and above. Try updating to a newer python version or "
                                   "replacing the lambda with a regular function.")
        return self._src


class _LambdaSrcFinder(ast.NodeVisitor):
    """
    Given a lambda function, isolate the lambda function source from any surrounding code.
    """

    def __init__(self, func):
        super().__init__()
        self.src = None
        # note that inspect.getsource gives the source for the line that contains the lambda
        # function, so we have to isolate the lambda function itself
        self.visit(ast.parse(textwrap.dedent(inspect.getsource(func)), filename='<string>'))

    def visit_Lambda(self, node):
        if self.src is not None:
            # it's possible to have multiple lambdas defined on the same line, so raise an error
            # if we find more than one.
            raise RuntimeError("Only one lambda function is allowed per line when using "
                               "_LambdaSrcFinder.")
        try:
            self.src = ast.unparse(node)
        except AttributeError:
            # ast.unparse was added in python 3.9
            self.src = None


if __name__ == '__main__':
    import pprint
    pprint.pprint(get_class_attributes(__file__))



import numpy as np
from scipy.sparse import coo_matrix
from scipy.sparse.linalg import splu

import openmdao.api as om


class MultiStatesComp(om.ImplicitComponent):

    def initialize(self):
        self.options.declare('num_elements', types=int)
        self.options.declare('force_vector', types=np.ndarray)
        self.options.declare('num_rhs', types=int)

    def setup(self):
        num_elements = self.options['num_elements']
        num_nodes = num_elements + 1
        size = 2 * num_nodes + 2
        num_rhs = self.options['num_rhs']

        self.add_input('K_local', shape=(num_elements, 4, 4))
        for j in range(num_rhs):
            self.add_output('d_%d' % j, shape=size)

        cols = np.arange(16*num_elements)
        rows = np.repeat(np.arange(4), 4)
        rows = np.tile(rows, num_elements) + np.repeat(np.arange(num_elements), 16) * 2

        for j in range(num_rhs):
            disp = 'd_%d' % j

            self.declare_partials(disp, 'K_local', rows=rows, cols=cols)
            self.declare_partials(disp, disp)

    def apply_nonlinear(self, inputs, outputs, residuals):
        num_rhs = self.options['num_rhs']

        self.K = self.assemble_CSC_K(inputs)
        for j in range(num_rhs):
            force_vector = np.concatenate([self.options['force_vector'][:, j], np.zeros(2)])
            residuals['d_%d' % j] = self.K.dot(outputs['d_%d' % j]) - force_vector

    def solve_nonlinear(self, inputs, outputs):
        num_rhs = self.options['num_rhs']

        self.K = self.assemble_CSC_K(inputs)
        self.lu = splu(self.K)

        for j in range(num_rhs):

            force_vector = np.concatenate([self.options['force_vector'][:, j], np.zeros(2)])
            outputs['d_%d' % j] = self.lu.solve(force_vector)

    def linearize(self, inputs, outputs, jacobian):
        num_rhs = self.options['num_rhs']
        num_elements = self.options['num_elements']

        self.K = self.assemble_CSC_K(inputs)
        self.lu = splu(self.K)

        i_elem = np.tile(np.arange(4), 4)
        i_d = np.tile(i_elem, num_elements) + np.repeat(np.arange(num_elements), 16) * 2

        K_dense = self.K.toarray()

        for j in range(num_rhs):
            disp = 'd_%d' % j
            jacobian[disp, 'K_local'] = outputs[disp][i_d]
            jacobian[disp, disp] = K_dense

    def solve_linear(self, d_outputs, d_residuals, mode):
        num_rhs = self.options['num_rhs']

        for j in range(num_rhs):
            disp = 'd_%d' % j

            if mode == 'fwd':
                d_outputs[disp] = self.lu.solve(d_residuals[disp])
            else:
                d_residuals[disp] = self.lu.solve(d_outputs[disp])

    def assemble_CSC_K(self, inputs):
        """
        Assemble the stiffness matrix in sparse CSC format.

        Returns
        -------
        ndarray
            Stiffness matrix as dense ndarray.
        """
        num_elements = self.options['num_elements']
        num_nodes = num_elements + 1
        num_entry = num_elements * 12 + 4
        ndim = num_entry + 4

        data = np.zeros((ndim, ), dtype=inputs._get_data().dtype)
        cols = np.empty((ndim, ))
        rows = np.empty((ndim, ))

        # First element.
        data[:16] = inputs['K_local'][0, :, :].flat
        cols[:16] = np.tile(np.arange(4), 4)
        rows[:16] = np.repeat(np.arange(4), 4)

        j = 16
        for ind in range(1, num_elements):
            ind1 = 2 * ind
            K = inputs['K_local'][ind, :, :]

            # NW quadrant gets summed with previous connected element.
            data[j-6:j-4] += K[0, :2]
            data[j-2:j] += K[1, :2]

            # NE quadrant
            data[j:j+4] = K[:2, 2:].flat
            rows[j:j+4] = np.array([ind1, ind1, ind1 + 1, ind1 + 1])
            cols[j:j+4] = np.array([ind1 + 2, ind1 + 3, ind1 + 2, ind1 + 3])

            # SE and SW quadrants together
            data[j+4:j+12] = K[2:, :].flat
            rows[j+4:j+12] = np.repeat(np.arange(ind1 + 2, ind1 + 4), 4)
            cols[j+4:j+12] = np.tile(np.arange(ind1, ind1 + 4), 2)

            j += 12

        data[-4:] = 1.0
        rows[-4] = 2 * num_nodes
        rows[-3] = 2 * num_nodes + 1
        rows[-2] = 0.0
        rows[-1] = 1.0
        cols[-4] = 0.0
        cols[-3] = 1.0
        cols[-2] = 2 * num_nodes
        cols[-1] = 2 * num_nodes + 1

        n_K = 2 * num_nodes + 2
        return coo_matrix((data, (rows, cols)), shape=(n_K, n_K)).tocsc()



"""
Various utilities for working with iterators.
"""


def meta2item_iter(metaiter, item):
    """
    Convert a metadata iterator to an iterator over (name, <item>).

    Parameters
    ----------
    metaiter : iter of (name, meta)
        Iterator over variable names and their metadata dicts.
    item : str
        The item to extract from the metadata.

    Yields
    ------
    tuple
        Tuple of (name, <item>) for each variable.
    """
    for name, meta in metaiter:
        yield name, meta[item]


def meta2items_iter(metaiter, items):
    """
    Convert a metadata iterator to an iterator over [name, meta[item0], meta[item1], ...].

    Parameters
    ----------
    metaiter : iter of (name, meta)
        Iterator over variable names and their metadata dicts.
    items : list of str
        The items to extract from the metadata.

    Yields
    ------
    list
        [name, meta[item0], meta[item1], ...] for each variable.
    """
    for name, meta in metaiter:
        toyield = [name]
        for item in items:
            toyield.append(meta[item])
        yield toyield


def size2range_iter(size_iter):
    """
    Convert a size iterator to a range iterator.

    Parameters
    ----------
    size_iter : iter of (name, size)
        Iterator over variable names and their sizes.

    Yields
    ------
    tuple
        Tuple of (name, (start, end)) for each variable.
    """
    start = 0
    for name, size in size_iter:
        yield name, (start, start + size)
        start += size



import unittest

import numpy as np
import openmdao.api as om

from openmdao.utils.mpi import MPI
from openmdao.utils.array_utils import evenly_distrib_idxs
from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.utils.assert_utils import assert_near_equal


if MPI:
    try:
        from openmdao.vectors.petsc_vector import PETScVector
    except ImportError:
        PETScVector = None


# check that pyoptsparse is installed. if it is, try to use SLSQP.
OPT, OPTIMIZER = set_pyoptsparse_opt('SLSQP')


@unittest.skipUnless(MPI and PETScVector, "MPI and PETSc are required.")
class ProbRemoteTestCase(unittest.TestCase):

    N_PROCS = 2

    def test_remote_var_access(self):
        # build the model
        prob = om.Problem()

        group = prob.model.add_subsystem('group', om.ParallelGroup())

        comp = om.ExecComp('f = (x-3)**2 + x*y + (y+4)**2 - 3', y=2.0)
        group.add_subsystem('comp1', comp)

        comp = om.ExecComp('g = x*y', y=2.0)
        group.add_subsystem('comp2', comp)

        prob.setup()

        prob['group.comp1.x'] = 4.
        prob['group.comp2.x'] = 5.

        prob.run_model()

        np.testing.assert_almost_equal(prob.get_val('group.comp1.f', get_remote=True), 42., decimal=5)
        np.testing.assert_almost_equal(prob.get_val('group.comp2.g', get_remote=True), 10., decimal=5)

    def test_remote_var_access_prom(self):
        prob = om.Problem()

        group = prob.model.add_subsystem('group', om.ParallelGroup(), promotes=['f', 'g'])

        group.add_subsystem('indep1', om.IndepVarComp('f'), promotes=['*'])
        group.add_subsystem('indep2', om.IndepVarComp('g'), promotes=['*'])

        prob.model.add_subsystem('summ', om.ExecComp('z = f + g'), promotes=['f', 'g'])
        prob.model.add_subsystem('prod', om.ExecComp('z = f * g'), promotes=['f', 'g'])

        prob.setup()

        prob['f'] = 4.
        prob['g'] = 5.

        prob.run_model()

        np.testing.assert_almost_equal(prob['summ.z'], 9., decimal=5)
        np.testing.assert_almost_equal(prob['prod.z'], 20., decimal=5)

    def test_is_local(self):
        p = om.Problem()
        p.model.add_subsystem('indep', om.IndepVarComp('x', 1.0))
        par = p.model.add_subsystem('par', om.ParallelGroup())
        par.add_subsystem('C1', om.ExecComp('y=2*x'))
        par.add_subsystem('C2', om.ExecComp('y=3*x'))
        p.model.connect('indep.x', ['par.C1.x', 'par.C2.x'])

        with self.assertRaises(RuntimeError) as err:
            p.is_local('indep.x')
        self.assertEqual(str(err.exception),
                         f"Problem {p._get_inst_id()}: is_local('indep.x') "
                         "was called before setup() completed.")

        with self.assertRaises(RuntimeError) as err:
            p.is_local('par.C1')
        self.assertEqual(str(err.exception),
                         f"Problem {p._get_inst_id()}: is_local('par.C1') "
                         "was called before setup() completed.")

        with self.assertRaises(RuntimeError) as err:
            p.is_local('par.C1.y')
        self.assertEqual(str(err.exception),
                         f"Problem {p._get_inst_id()}: is_local('par.C1.y') "
                         "was called before setup() completed.")

        with self.assertRaises(RuntimeError) as err:
            p.is_local('par.C1.x')
        self.assertEqual(str(err.exception),
                         f"Problem {p._get_inst_id()}: is_local('par.C1.x') "
                         "was called before setup() completed.")

        p.setup()
        p.final_setup()

        self.assertTrue(p.is_local('indep'), 'indep should be local')
        self.assertTrue(p.is_local('indep.x'), 'indep.x should be local')

        if p.comm.rank == 0:
            self.assertTrue(p.is_local('par.C1'), 'par.C1 should be local')
            self.assertTrue(p.is_local('par.C1.x'), 'par.C1.x should be local')
            self.assertTrue(p.is_local('par.C1.y'), 'par.C1.y should be local')

            self.assertFalse(p.is_local('par.C2'), 'par.C1 should be remote')
            self.assertFalse(p.is_local('par.C2.x'), 'par.C1.x should be remote')
            self.assertFalse(p.is_local('par.C2.y'), 'par.C1.y should be remote')
        else:
            self.assertFalse(p.is_local('par.C1'), 'par.C1 should be remote')
            self.assertFalse(p.is_local('par.C1.x'), 'par.C1.x should be remote')
            self.assertFalse(p.is_local('par.C1.y'), 'par.C1.y should be remote')

            self.assertTrue(p.is_local('par.C2'), 'par.C2 should be local')
            self.assertTrue(p.is_local('par.C2.x'), 'par.C2.x should be local')
            self.assertTrue(p.is_local('par.C2.y'), 'par.C2.y should be local')

    def test_get_remote(self):

        N = 3

        class DistribComp(om.ExplicitComponent):

            def setup(self):
                rank = self.comm.rank
                sizes, offsets = evenly_distrib_idxs(self.comm.size, N)

                self.add_input('x', shape=1, distributed=True)
                self.add_output('y', shape=sizes[rank], distributed=True)

            def compute(self, inputs, outputs):
                rank = self.comm.rank
                sizes, offsets = evenly_distrib_idxs(self.comm.size, N)

                outputs['y'] = inputs['x']*np.ones((sizes[rank],))
                if rank == 0:
                    outputs['y'][0] = 2.

        class MyModel(om.Group):

            def setup(self):
                self.add_subsystem('ivc', om.IndepVarComp('x', 0.))
                self.add_subsystem('dst', DistribComp())
                self.add_subsystem('sum', om.ExecComp('z = sum(y)', y=np.zeros((N,)), z=0.0))

                self.connect('ivc.x', 'dst.x', src_indices=[0])
                self.connect('dst.y', 'sum.y', src_indices=om.slicer[:])

                self.add_subsystem('par', om.ParallelGroup())
                self.par.add_subsystem('c1', om.ExecComp(['y=2.0*x']), promotes_inputs=['*'])
                self.par.add_subsystem('c2', om.ExecComp(['y=5.0*x']), promotes_inputs=['*'])

                self.connect('ivc.x', 'par.x')

        prob = om.Problem(model=MyModel())
        prob.setup(mode='fwd')

        prob['ivc.x'] = 7.0
        prob.run_model()

        # get_remote=True
        assert_near_equal(prob.get_val('ivc.x', get_remote=True), [7.])
        assert_near_equal(prob.get_val('dst.x', get_remote=True), [7., 7.])  # ???????
        assert_near_equal(prob.get_val('dst.y', get_remote=True), [2., 7., 7.])
        assert_near_equal(prob.get_val('par.c1.x', get_remote=True), [7.])
        assert_near_equal(prob.get_val('par.c1.y', get_remote=True), [14.])
        assert_near_equal(prob.get_val('par.c2.x', get_remote=True), [7.])
        assert_near_equal(prob.get_val('par.c2.y', get_remote=True), [35.])

        self.maxDiff = None

        remote_msg = ("<model> <class MyModel>: Variable '{name}' is not local to rank {rank}. "
                      "You can retrieve values from other processes using "
                      "`get_val(<name>, get_remote=True)`.")

        distrib_msg = ("<model> <class MyModel>: Variable '{name}' is a distributed variable. "
                       "You can retrieve values from all processes using "
                       "`get_val(<name>, get_remote=True)` or from the local "
                       "process using `get_val(<name>, get_remote=False)`.")

        if prob.comm.rank == 0:
            #
            # get_remote=False
            #

            # get the local part of the distributed inputs/outputs
            assert_near_equal(prob.get_val('dst.x', get_remote=False), [7.])  # from src ('ivc.x')
            assert_near_equal(prob.model.get_val('dst.x', get_remote=False, from_src=False), [7.])
            assert_near_equal(prob.get_val('dst.y', get_remote=False), [2., 7.])

            # par.c1 is local
            assert_near_equal(prob.get_val('par.c1.x', get_remote=False), [7.])  # from src ('ivc.x')
            assert_near_equal(prob.model.get_val('par.c1.x', get_remote=False, from_src=False), [7.])
            assert_near_equal(prob.get_val('par.c1.y', get_remote=False), [14.])

            # par.c2 is remote
            assert_near_equal(prob.get_val('par.c2.x', get_remote=False), [7.])  # from src ('ivc.x')

            with self.assertRaises(RuntimeError) as cm:
                prob.model.get_val('par.c2.x', get_remote=False, from_src=False)
            self.assertEqual(str(cm.exception), remote_msg.format(name='par.c2.x', rank=0))

            with self.assertRaises(RuntimeError) as cm:
                prob.get_val('par.c2.y', get_remote=False)
            self.assertEqual(str(cm.exception), remote_msg.format(name='par.c2.y', rank=0))

            #
            # get_remote=None
            #

            with self.assertRaises(RuntimeError) as cm:
                prob['dst.x']
            self.assertEqual(str(cm.exception), distrib_msg.format(name='dst.x'))

            with self.assertRaises(RuntimeError) as cm:
                prob.model.get_val('dst.x', get_remote=None, from_src=False)
            self.assertEqual(str(cm.exception), distrib_msg.format(name='dst.x'))

            with self.assertRaises(RuntimeError) as cm:
                prob['dst.y']
            self.assertEqual(str(cm.exception), distrib_msg.format(name='dst.y'))

            # par.c1 is local
            assert_near_equal(prob['par.c1.x'], [7.])  # from src ('ivc.x')
            assert_near_equal(prob.model.get_val('par.c1.x', get_remote=None, from_src=False), [7.])
            assert_near_equal(prob['par.c1.y'], [14.])

            # par.c2 is remote
            assert_near_equal(prob['par.c2.x'], [7.])  # from src ('ivc.x')

            with self.assertRaises(RuntimeError) as cm:
                prob.model.get_val('par.c2.x', get_remote=None, from_src=False)
            self.assertEqual(str(cm.exception), remote_msg.format(name='par.c2.x', rank=0))

            with self.assertRaises(RuntimeError) as cm:
                prob['par.c2.y']
            self.assertEqual(str(cm.exception), remote_msg.format(name='par.c2.y', rank=0))

        else:
            #
            # get_remote=False
            #

            # get the local part of the distributed inputs/outputs
            assert_near_equal(prob.get_val('dst.x', get_remote=False), [7.])  # from src ('ivc.x')
            assert_near_equal(prob.model.get_val('dst.x', get_remote=False, from_src=False), [7.])
            assert_near_equal(prob.get_val('dst.y', get_remote=False), [7.])

            # par.c1 is remote
            prob.get_val('par.c1.x', get_remote=False)  # from src ('ivc.x')

            with self.assertRaises(RuntimeError) as cm:
                prob.model.get_val('par.c1.x', get_remote=False, from_src=False)
            self.assertEqual(str(cm.exception), remote_msg.format(name='par.c1.x', rank=1))

            with self.assertRaises(RuntimeError) as cm:
                prob.get_val('par.c1.y', get_remote=False)
            self.assertEqual(str(cm.exception), remote_msg.format(name='par.c1.y', rank=1))

            # par.c2 is local
            assert_near_equal(prob.get_val('par.c2.x', get_remote=False), [7.])  # from src ('ivc.x')
            assert_near_equal(prob.model.get_val('par.c2.x', get_remote=False, from_src=False), [7.])
            assert_near_equal(prob.get_val('par.c2.y', get_remote=False), [35.])

            #
            # get_remote=None
            #

            with self.assertRaises(RuntimeError) as cm:
                prob['dst.x']
            self.assertEqual(str(cm.exception), distrib_msg.format(name='dst.x'))

            with self.assertRaises(RuntimeError) as cm:
                prob.model.get_val('dst.x', get_remote=None, from_src=False)
            self.assertEqual(str(cm.exception), distrib_msg.format(name='dst.x'))

            with self.assertRaises(RuntimeError) as cm:
                prob['dst.y']
            self.assertEqual(str(cm.exception), distrib_msg.format(name='dst.y'))

            # par.c1 is remote
            assert_near_equal(prob['par.c1.x'], [7.])  # from src ('ivc.x')

            with self.assertRaises(RuntimeError) as cm:
                prob.model.get_val('par.c1.x', get_remote=None, from_src=False)
            self.assertEqual(str(cm.exception), remote_msg.format(name='par.c1.x', rank=1))

            with self.assertRaises(RuntimeError) as cm:
                prob['par.c1.y']
            self.assertEqual(str(cm.exception), remote_msg.format(name='par.c1.y', rank=1))

            # par.c2 is local
            assert_near_equal(prob['par.c2.x'], [7.])  # from src ('ivc.x')
            assert_near_equal(prob.model.get_val('par.c2.x', get_remote=None, from_src=False), [7.])
            assert_near_equal(prob['par.c2.y'], [35.])


@unittest.skipUnless(MPI and PETScVector, "MPI and PETSc are required.")
@unittest.skipIf(OPT is None or OPTIMIZER is None, "pyOptSparse is required.")
class ProbRemote4TestCase(unittest.TestCase):

    N_PROCS = 4

    def test_prob_split_comm(self):
        colors = [0, 0, 1, 1]
        comm = MPI.COMM_WORLD.Split(colors[MPI.COMM_WORLD.rank])

        # split the size 4 comm into 2 size 2 comms
        self.assertEqual(comm.size, 2)

        prob = om.Problem(comm=comm)
        model = prob.model

        p1 = model.add_subsystem('p1', om.IndepVarComp('x', 99.0))
        p1.add_design_var('x', lower=-50.0, upper=50.0)

        par = model.add_subsystem('par', om.ParallelGroup())
        par.add_subsystem('C1', om.ExecComp('y = x*x'))
        par.add_subsystem('C2', om.ExecComp('y = x*x'))

        model.add_subsystem('obj', om.ExecComp('o = a + b + 2.'))

        model.connect('p1.x', ['par.C1.x', 'par.C2.x'])
        model.connect('par.C1.y', 'obj.a')
        model.connect('par.C2.y', 'obj.b')

        model.add_objective('obj.o')

        prob.set_solver_print(level=0)

        prob.driver = om.pyOptSparseDriver()
        prob.driver.options['optimizer'] = OPTIMIZER
        prob.driver.options['print_results'] = False

        prob.setup()
        prob.run_model()

        failed = not prob.run_driver().success

        all_failed = comm.allgather(failed)
        if any(all_failed):
            msg = 'No solution found' if prob.driver.pyopt_solution is None \
                else str(prob.driver.pyopt_solution.optInform)
            all_msgs = comm.allgather(msg)
            for i, tup in enumerate(zip(all_failed, all_msgs)):
                failed, msg = tup
                if failed:
                    self.fail(f"Optimization failed on rank {i}: {msg}")

        try:
            obj = prob['obj.o']
            failed = False
        except Exception as err:
            obj = str(err)
            failed = True

        all_status = comm.allgather((obj, failed))
        for i, tup in enumerate(all_status):
            obj, failed = tup
            if failed:
                self.fail(f"Failed to retrieve objective on rank {i}: {obj}")
            assert_near_equal(obj, 2.0, 1e-6)



"""Define the LinearBlockJac class."""
from openmdao.solvers.solver import BlockLinearSolver


class LinearBlockJac(BlockLinearSolver):
    """
    Linear block Jacobi solver.

    Parameters
    ----------
    **kwargs : dict
        Options dictionary.
    """

    SOLVER = 'LN: LNBJ'

    def _single_iteration(self):
        """
        Perform the operations in the iteration loop.
        """
        system = self._system()
        mode = self._mode

        subs = list(system._relevance.filter(system._subsystems_myproc))
        scopelist = [None] * len(subs)

        if mode == 'fwd':
            system._transfer('linear', mode)

            for i, subsys in enumerate(subs):
                scope_out, scope_in = system._get_matvec_scope(subsys)
                scope_out = self._vars_union(self._scope_out, scope_out)
                scope_in = self._vars_union(self._scope_in, scope_in)
                scopelist[i] = (scope_out, scope_in)
                if subsys._iter_call_apply_linear():
                    subsys._apply_linear(None, mode, scope_out, scope_in)
                else:
                    subsys._dresiduals.set_val(0.0)

            system._dresiduals *= -1.0
            system._dresiduals += self._rhs_vec

            for i, subsys in enumerate(subs):
                scope_out, scope_in = scopelist[i]
                subsys._solve_linear(mode, scope_out, scope_in)

        else:  # rev
            for i, subsys in enumerate(subs):
                scope_out, scope_in = system._get_matvec_scope(subsys)
                scope_out = self._vars_union(self._scope_out, scope_out)
                scope_in = self._vars_union(self._scope_in, scope_in)
                scopelist[i] = (scope_out, scope_in)
                if subsys._iter_call_apply_linear():
                    subsys._apply_linear(None, mode, scope_out, scope_in)
                else:
                    subsys._doutputs.set_val(0.0)

            system._transfer('linear', mode)

            system._doutputs *= -1.0
            system._doutputs += self._rhs_vec

            for i, subsys in enumerate(subs):
                scope_out, scope_in = scopelist[i]
                subsys._solve_linear(mode, scope_out, scope_in)



"""
Utilities for working with files.
"""
import sys
import os
import importlib
import types
from collections.abc import Iterable
from fnmatch import fnmatch
from os.path import join, basename, dirname, isfile, split, splitext, abspath
import pathlib
import shutil

from openmdao.utils.om_warnings import issue_warning


def get_module_path(fpath):
    """
    Given a module filename, return its full Python module path.

    This includes enclosing packages and is based on existence of ``__init__.py`` files.

    Parameters
    ----------
    fpath : str
        Pathname of file.

    Returns
    -------
    str or None
        Full module path of the given file.  Returns None if the file is not part of a package.
    """
    fpath = abspath(fpath)
    if basename(fpath).startswith('__init__.'):
        pnames = []
    else:
        pnames = [splitext(basename(fpath))[0]]
    path = dirname(fpath)

    initfile = join(path, '__init__.py')
    if not isfile(initfile):
        return None

    while isfile(initfile):
        path, pname = split(path)
        pnames.append(pname)
        initfile = join(path, '__init__.py')

    return '.'.join(pnames[::-1])


def package_iter(start_dir='.', dir_includes=None, dir_excludes=(), file_includes=None,
                 file_excludes=()):
    """
    Iterate over python files in packages (recursively) starting in start_dir.

    NOTE : all *_includes and *_excludes are applied to *local* directory and file names.

    Parameters
    ----------
    start_dir : str
        Starting directory.
    dir_includes : iter of str or None
        Glob patterns for directory inclusion. Be careful here because dir names are local,
        so, for example, if includes=('foo',), then directory 'foo' would be included, but
        any subdirectories of 'foo' that were not also named 'foo' would not.
    dir_excludes : iter of str
        Glob patterns for directory exclusion.
    file_includes : iter of str or None
        Glob patterns for file inclusion.
    file_excludes : iter of str
        Glob patterns for file exclusion.

    Yields
    ------
    str
        Filenames (full path from start_dir).
    """
    file_includes = set() if file_includes is None else set(file_includes)
    file_includes.add('*.py')
    file_excludes = set() if file_excludes is None else set(file_excludes)
    file_excludes.update(('_*', 'test_*', 'api.py', 'parallel_api.py'))
    dir_excludes = set() if dir_excludes is None else set(dir_excludes)
    dir_excludes.update(('test', 'tests'))

    for f in files_iter(start_dir, dir_includes=dir_includes, dir_excludes=dir_excludes,
                        file_includes=file_includes, file_excludes=file_excludes,
                        package_only=True):
        yield f


def files_iter(start_dir='.', dir_includes=None, dir_excludes=(),
               file_includes=None, file_excludes=(), package_only=False):
    """
    Iterate over files (recursively) starting in start_dir.

    NOTE : all *_includes and *_excludes are applied to *local* directory and file names.

    Parameters
    ----------
    start_dir : str
        Starting directory.
    dir_includes : iter of str or None
        Glob patterns for directory inclusion. Be careful here because dir names are local,
        so, for example, if includes=('foo',), then directory 'foo' would be included, but
        any subdirectories of 'foo' that were not also named 'foo' would not.
    dir_excludes : iter of str
        Glob patterns for directory exclusion.
    file_includes : iter of str or None
        Glob patterns for file inclusion.
    file_excludes : iter of str
        Glob patterns for file exclusion.
    package_only : bool
        If True, only yield files that are contained in a python package.

    Yields
    ------
    str
        Filenames (full path from start_dir).
    """
    for root, dirs, files in os.walk(start_dir):
        if package_only and '__init__.py' not in files:
            dirs[:] = []
            continue
        for pat in dir_excludes:
            dirs[:] = sorted([d for d in dirs if not fnmatch(d, pat)])
        if dir_includes:
            incdirs = set()
            for pat in dir_includes:
                incdirs.update(d for d in dirs if fnmatch(d, pat))
            dirs[:] = sorted(incdirs)
        for f in files:
            for pat in file_excludes:
                if fnmatch(f, pat):
                    break
            else:
                if file_includes:
                    for pat in file_includes:
                        if fnmatch(f, pat):
                            yield join(root, f)
                else:
                    yield join(root, f)


def _to_filename(spec):
    """
    Return the filename part of the given testspec or the full string if the string is a filename.

    Parameters
    ----------
    spec : str
        The filename or testspec.

    Returns
    -------
    str
        The filename.
    """
    if ':' in spec and not isfile(spec):
        fname, _ = spec.rsplit(':', 1)
        if not fname.endswith('.py'):
            try:
                mod = importlib.import_module(fname)
                return mod.__file__
            except ImportError:
                return spec
        return fname

    return spec


def _load_and_exec(script_name, user_args):
    """
    Load and exec the given script as __main__.

    Parameters
    ----------
    script_name : str
        The name of the script to load and exec.
    user_args : list of str
        Args to be passed to the user script.
    """
    if ':' in script_name and not isfile(script_name):
        return _load_and_run_test(script_name)

    sys.path.insert(0, dirname(script_name))

    sys.argv[:] = [script_name] + user_args

    with open(script_name, 'rb') as fp:
        code = compile(fp.read(), script_name, 'exec')

    globals_dict = {
        '__file__': script_name,
        '__name__': '__main__',
        '__package__': None,
        '__cached__': None,
    }

    exec(code, globals_dict)  # nosec: private, internal use only


def fname2mod_name(fname):
    """
    Convert a string to a valid python module name.

    Parameters
    ----------
    fname : str
        The filename to convert.

    Returns
    -------
    str
        A valid module name corresponding to the given filename.
    """
    to_replace = ['-', ' ', '.', '(', ')', '[', ']', '{', '}', '=', '+'
                  '!', '@', '#', '$', '%', '^', '&', '*', '~', '`',
                  ';', ':', '"', "'", '<', '>', '?', '/', '\\', '|']

    if not fname.endswith('.py'):
        raise ValueError(f"'{fname}' does not end with '.py'")

    s = basename(fname).rsplit('.', 1)[0]

    for c in to_replace:
        s = s.replace(c, '_')
    return s


def _load_and_run_test(testspec):
    """
    Load and run an individual test function.

    Parameters
    ----------
    testspec : str
        <fpath_or_modpath>:<testcase>.<method> OR <fpath_or_modpath>:<function>
    """
    syspath_save = sys.path[:]

    modpath, funcpath = testspec.rsplit(':', 1)
    orig_modpath = modpath

    if modpath.endswith('.py'):
        modpath = get_module_path(modpath)
        if modpath is None:
            # create a module dynamically
            modpath = fname2mod_name(orig_modpath)
            mod = types.ModuleType(modpath)
            sys.modules[modpath] = mod
            mod.__file__ = modpath
            mod.__name__ = modpath
            mod.__package__ = None
            mod.__cached__ = None
            with open(orig_modpath, 'rb') as fp:
                code = compile(fp.read(), orig_modpath, 'exec')
            exec(code, mod.__dict__)  # nosec: private, internal use only
        else:
            mod = importlib.import_module(modpath)
    else:
        mod = importlib.import_module(modpath)

    try:
        return _run_test_func(mod, funcpath)
    finally:
        sys.path = syspath_save


def _run_test_func(mod, funcpath):
    """
    Run the given TestCase method or test function in the given module.

    Parameters
    ----------
    mod : module
        The module where the test resides.
    funcpath : str
        Either <testcase>.<method_name> or <func_name>.

    Returns
    -------
    object
        In the case of a module level function call, returns whatever the function returns.
    """
    parts = funcpath.split('.', 1)
    if len(parts) == 2:
        tcase_name, method_name = parts
        testcase = getattr(mod, tcase_name)(methodName=method_name)
        setup = getattr(testcase, 'setUp', None)
        if setup is not None:
            setup()
        getattr(testcase, method_name)()
        teardown = getattr(testcase, 'tearDown', None)
        if teardown:
            teardown()
    else:
        funcname = parts[0]
        return getattr(mod, funcname)()


if sys.version_info >= (3, 8):
    from importlib.metadata import entry_points

    if sys.version_info >= (3, 10):
        def _eps_get(group):
            eps = entry_points().select(group=group)
            for name in eps.names:
                yield eps[name]
    else:
        def _eps_get(group):
            eps = entry_points()
            if group in eps:
                yield from eps[group]

    def _iter_entry_points(group):
        # there seems to be a bug currently where entry points can show up more than
        # once in the iterator, so keep track of the ones we've already seen.
        # TODO: revisit later to see if we can remove the check
        seen = set()
        for ep in _eps_get(group):
            if ep.name not in seen:
                seen.add(ep.name)
                yield ep
else:
    try:
        import pkg_resources
    except ImportError:
        def _iter_entry_points(group):
            issue_warning("Can't retrieve entry points because pkg_resources is not installed. "
                          "Either install it using 'pip install setuptools' or upgrade to python "
                          "3.8 or newer.")
            return ()
    else:
        def _iter_entry_points(group):
            yield from pkg_resources.iter_entry_points(group)


def text2html(text, title='', style=None):
    """
    Wrap the given text for display as an html file.

    Returns an html syntax string that can be written to a file.

    Parameters
    ----------
    text : str
        Text to be displayed.
    title : str
        Title to display above text.
    style : str or None
        If not None, use as the contents of the style block for the enclosing <pre> tag.

    Returns
    -------
    str
        Content string to create an html file.
    """
    if style is None:
        style = """
            display: block;
            font-family: monospace;
            font-size: 1.5em;
            white-space: pre;
            margin: 1em 0;
        """
    return """
<!DOCTYPE html>
<html lang="en">
<head>
    <style>
        .center {
            display: block;
            margin-left: auto;
            margin-right: auto;
            width: 90%;
        }
        h2 {text-align: center;}
        pre {""" + style + """
        }
    </style>
</head>
<body>
<h2>""" + title + """</h2>
<pre>
""" + text + """
</pre>
</body>
</html>
"""


def image2html(imagefile, title='', alt=''):
    """
    Wrap the given image for display as an html file.

    Returns an html syntax string that can be written to a file.

    Parameters
    ----------
    imagefile : str
        Name of image file to be displayed.
    title : str
        The page title.
    alt : str
        Set the alt text for the image.

    Returns
    -------
    str
        Content string to create an html file.
    """
    return """
<!DOCTYPE html>
<html lang="en">
<head>
    <style>
        h2 {text-align: center;}
        .center {
            display: block;
            margin-left: auto;
            margin-right: auto;
            width: 80%;
        }
    </style>
</head>
<body>
<h2>""" + title + "</h2>" + f"""
<img src="{imagefile}" alt="{alt}" class="center"></img>

</body>
</html>
"""


def get_work_dir():
    """
    Return either os.getcwd() or the value of the OPENMDAO_WORKDIR environment variable.

    Returns
    -------
    str
        The working directory.
    """
    workdir = os.environ.get('OPENMDAO_WORKDIR', '')

    if workdir:
        return workdir

    return os.getcwd()


def _get_outputs_dir(obj, *subdirs, mkdir=True):
    """
    Return a pathlib.Path for the outputs directory related to the given problem or system.

    This path is based on the "problem path" in a hierarchy of problems.
    The resulting outputs directory will be nested where each problem's output directory
    contains its own output files and subdirectories as well as any subproblems.

    This directory also will include a .openmdao_outputs hidden file that
    marks this directory as being created by OpenMDAO. This makes identifying the
    directory during cleanup more reliable.

    Parameters
    ----------
    obj : Problem or System or Solver
        The problem or system or Solver from which we are opening a file.
    subdirs : str
        Additional subdirectories under the top level directory for the relevant problem. Each
        subdir is passed as a separate positional argument.
    mkdir : bool
        If True, force the creation of this directory.
    """
    from openmdao.core.problem import Problem
    from openmdao.core.system import System
    from openmdao.solvers.solver import Solver

    if isinstance(obj, Problem):
        prob_meta = obj._metadata
    elif isinstance(obj, System):
        prob_meta = obj._problem_meta
    elif isinstance(obj, Solver):
        system = obj._system
        if system is None:
            raise RuntimeError('The output directory for Solvers cannot be accessed '
                               'before final_setup.')
        prob_meta = system()._problem_meta
    else:
        raise RuntimeError(f'Cannot get problem metadata for object: {obj}')

    if prob_meta is None or prob_meta.get('pathname', None) is None:
        raise RuntimeError('The output directory cannot be accessed before setup.')

    prob_pathname = prob_meta['pathname']

    work_dir = pathlib.Path(get_work_dir())
    if mkdir and not work_dir.exists():
        work_dir.mkdir(exist_ok=True)

    outs_dir = work_dir

    # it's possible that a sub-problem requests the output directory before its parent problem,
    # so we need to check existence of the parent directories for all parent problems and create
    # them (ensuring the .openmdao_out file is present) if they don't exist.  Otherwise they
    # won't be properly identified during cleanup.

    # Also, we don't check if rank==0 here because when mkdir is True, we need to ensure that the
    # directory has been created by the time we return, so just letting any rank create the file
    # and handling race conditions by using exist_ok=True works better than trying to limit creation
    # to rank==0 and performing some sort of barrier operation to ensure that the file is created
    # before returning in all ranks.
    for p in prob_pathname.split('/'):
        outs_dir = outs_dir / f'{p}_out'
        if mkdir and not outs_dir.exists():
            outs_dir.mkdir(exist_ok=True)

            # Touch the .openmdao_out file for the output directory to ease identification.
            outs_file = outs_dir / '.openmdao_out'
            if not outs_file.exists():
                try:
                    open(outs_file, 'w').close()
                except OSError:
                    pass

    if subdirs:
        dirpath = outs_dir / pathlib.Path(*subdirs)

        if mkdir and not dirpath.exists():
            dirpath.mkdir(parents=True, exist_ok=True)

        return dirpath

    return outs_dir


def _is_openmdao_output_dir(directory):
    """
    Check if a directory is an OpenMDAO output directory.

    Parameters
    ----------
    directory : str or Path
        The directory to check.

    Returns
    -------
    bool
        True if the directory is an OpenMDAO output directory, False otherwise.
    """
    directory = pathlib.Path(directory)
    return directory.is_dir() and (directory / '.openmdao_out').exists()


def _find_openmdao_output_dirs(paths, pattern='*_out', recurse=False):
    """
    Find all OpenMDAO output directories in the given path.

    Parameters
    ----------
    paths : str or Path or Iterable
        The path to search for OpenMDAO output directories.
    pattern : str
        A glob pattern that the output directories are required to match.
    recurse : bool
        If True, search recursively.

    Returns
    -------
    list
        A list of OpenMDAO output directories.
    """
    if isinstance(paths, (str, pathlib.Path)):
        paths = [paths]
    elif not isinstance(paths, Iterable):
        raise ValueError("The 'paths' parameter must be a string, Path, or an iterable of them.")

    openmdao_dirs = []
    for path in paths:
        path = pathlib.Path(path)
        if not path.is_dir():
            continue

        for root, dirs, _ in os.walk(path):
            # Use a copy of the dirs list to avoid modifying it while iterating
            root_path = pathlib.Path(root)
            if _is_openmdao_output_dir(root) and fnmatch(root_path.name, pattern):
                openmdao_dirs.append(root_path)
            for d in dirs[:]:
                dir_path = pathlib.Path(root) / d
                if _is_openmdao_output_dir(dir_path):
                    if fnmatch(dir_path.name, pattern):
                        openmdao_dirs.append(dir_path)
                    dirs.remove(d)  # Do not recurse into OpenMDAO output directories
            if not recurse:
                break
    return openmdao_dirs


def clean_outputs(obj='.', recurse=False, prompt=True, pattern='*_out', dryrun=False):
    """
    Remove output directories created by OpenMDAO.

    A directory is determined to be an OpenMDAO output directory if its name
    ends in `_out` and it contains the file `.openmdao_out`.

    Parameters
    ----------
    obj : Problem or System or Solver or str or Path
        The problem or system or solver whose output file should be removed.
    recurse : bool
        If True, and if obj is a string or Path, recurse into it
        finding and removing OpenMDAO output directories along the way.
        This option is ignored if obj is a Problem, System, or Solver.
    prompt : bool
        If True, prompt the user to confirm directories to be removed.
        This option is ignored if obj is a Problem, System, or Solver.
    pattern : str
        A glob pattern used for matching directories.
    dryrun : bool
        If True, report which directories would be removed without actually removing them.
    """
    output_dirs = []

    if isinstance(obj, (str, pathlib.Path)):
        # A single pathname or path object was given.
        output_dirs = _find_openmdao_output_dirs(obj, pattern=pattern, recurse=recurse)
    elif isinstance(obj, (Iterable,)):
        # Multiple paths given
        output_dirs.extend(_find_openmdao_output_dirs(obj, pattern, recurse))
    elif hasattr(obj, 'get_outputs_dir'):
        output_dir = obj.get_outputs_dir(mkdir=False)
        prompt = False
        if output_dir and _is_openmdao_output_dir(output_dir):
            output_dirs.append(pathlib.Path(output_dir))

    if not output_dirs:
        print('No OpenMDAO output directories found.')
        return
    else:
        print(f'Found {len(output_dirs)} OpenMDAO output directories:')

    removed_count = 0
    for dir_path in sorted(output_dirs):
        if dryrun:
            print(f'Would remove {dir_path} (dryrun = True).')
        elif prompt:
            response = input(f"Remove {dir_path}? [y/N] ").strip().lower()
            if response == 'y':
                shutil.rmtree(dir_path)
                print(f'Removed {dir_path}')
                removed_count += 1
        else:
            shutil.rmtree(dir_path)
            print(f'Removed {dir_path}')
            removed_count += 1

    print(f'Removed {removed_count} OpenMDAO output directories.')



import unittest
import sys
import itertools

import numpy as np
from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials, assert_check_totals
import openmdao.api as om

from openmdao.utils.jax_utils import jax, jnp
from openmdao.utils.testing_utils import parameterized_name

try:
    from parameterized import parameterized
except ImportError:
    from openmdao.utils.assert_utils import SkipParameterized as parameterized


class DotProd(om.ExplicitComponent):
    def setup(self):
        self.add_input('x', shape_by_conn=True)
        self.add_input('y', shape_by_conn=True)
        self.add_output('z', compute_shape=lambda shapes: (shapes['x'][0], shapes['y'][1]))

        self.declare_partials(of='z', wrt=['x', 'y'])

    def compute(self, inputs, outputs):
        outputs['z'] = np.dot(inputs['x'], inputs['y'])


class DotProdMult(om.ExplicitComponent):
    def setup(self):
        self.add_input('x', shape_by_conn=True)
        self.add_input('y', shape_by_conn=True)
        self.add_output('z', compute_shape=lambda shapes: (shapes['x'][0], shapes['y'][1]))
        self.add_output('zz', copy_shape='y')

        self.declare_partials(of=['z', 'zz'], wrt=['x', 'y'])

    def compute(self, inputs, outputs):
        outputs['z'] = np.dot(inputs['x'], inputs['y'])
        outputs['zz'] = inputs['y'] * 2.5


class DotProdMultPrimalNoDeclPartials(om.JaxExplicitComponent):
    def setup(self):
        self.add_input('x', shape_by_conn=True)
        self.add_input('y', shape_by_conn=True)
        self.add_output('z', compute_shape=lambda shapes: (shapes['x'][0], shapes['y'][1]))
        self.add_output('zz', copy_shape='y')

    def compute_primal(self, x, y):
        z = jnp.dot(x, y)
        zz = y * 2.5
        return z, zz


class DotProdMultPrimal(DotProdMultPrimalNoDeclPartials):
    def setup(self):
        super().setup()
        self.declare_partials(of=['z', 'zz'], wrt=['x', 'y'])


class DotProdMultPrimalOption(om.JaxExplicitComponent):
    def __init__(self, stat=2., **kwargs):
        super().__init__(**kwargs)
        self.stat = stat

    def initialize(self):
        self.options.declare('mult', default=1.0, types=float, allow_none=False)

    def setup(self):
        self.add_input('x', shape_by_conn=True)
        self.add_input('y', shape_by_conn=True)
        self.add_output('z', compute_shape=lambda shapes: (shapes['x'][0], shapes['y'][1]))
        self.add_output('zz', copy_shape='y')

        self.declare_partials(of=['z', 'zz'], wrt=['x', 'y'])

    def get_self_statics(self):
        return (self.options['mult'], self.stat)

    def compute_primal(self, x, y):
        z = jnp.dot(x, y)
        zz = y * self.options['mult'] * self.stat
        return z, zz



class DotProdMultOption(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('mult', default=2.5, desc='multiplier', types=(float,))

    def setup(self):
        self.add_input('x', shape_by_conn=True)
        self.add_input('y', shape_by_conn=True)
        self.add_output('z', compute_shape=lambda shapes: (shapes['x'][0], shapes['y'][1]))
        self.add_output('zz', copy_shape='y')

        self.declare_partials(of=['z', 'zz'], wrt=['x', 'y'])

    def compute(self, inputs, outputs):
        outputs['z'] = np.dot(inputs['x'], inputs['y'])
        outputs['zz'] = inputs['y'] * self.options['mult']



class DotProductMultDiscrete(om.ExplicitComponent):
    def setup(self):
        self.add_input('x', shape_by_conn=True)
        self.add_input('y', shape_by_conn=True)
        self.add_discrete_input('disc_in', val=2)
        self.add_output('z', compute_shape=lambda shapes: (shapes['x'][0], shapes['y'][1]))
        self.add_output('zz', copy_shape='y')
        self.add_discrete_output('disc_out', val=3)

        self.declare_partials(of=['z', 'zz'], wrt=['x', 'y'])

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        discrete_outputs['disc_out'] = -discrete_inputs['disc_in']
        if discrete_inputs['disc_in'] > 0:
            outputs['z'] = np.dot(inputs['x'], inputs['y'])
        else:
            outputs['z'] = -np.dot(inputs['x'], inputs['y'])

        if discrete_outputs['disc_out'] > 0:
            outputs['zz'] = inputs['y'] * 2.5
        else:
            outputs['zz'] = inputs['y'] * 3.0



class DotProductMultDiscretePrimal(om.ExplicitComponent):
    def setup(self):
        self.add_input('x', shape_by_conn=True)
        self.add_input('y', shape_by_conn=True)
        self.add_discrete_input('disc_in', val=2)
        self.add_output('z', compute_shape=lambda shapes: (shapes['x'][0], shapes['y'][1]))
        self.add_output('zz', copy_shape='y')
        self.add_discrete_output('disc_out', val=3)

        self.declare_partials(of=['z', 'zz'], wrt=['x', 'y'])

    def compute_primal(self, x, y, disc_in):
        def pos(x, y):
            return jnp.dot(x, y), y * 3.0

        def neg(x, y):
            return -jnp.dot(x, y), y * 2.5

        disc_out = -disc_in
        z, zz = jax.lax.cond(disc_in >= 0, pos, neg, x, y)

        self._discrete_outputs.set_vals((disc_out,))
        return (z, zz, disc_out)

x_shape = (2, 3)
y_shape = (3, 4)


@unittest.skipIf(jax is None or sys.version_info < (3, 9), 'jax is not available or python < 3.9.')
class TestJaxComp(unittest.TestCase):

    @parameterized.expand(itertools.product(['fwd', 'rev'], [(True, 'jax'), (False, 'jax'), (False, 'fd')]), name_func=parameterized_name)
    def test_jax_explicit_comp(self, mode, tup):
        matrix_free, derivs_method = tup
        xshape = x_shape
        yshape = y_shape
        p = om.Problem()
        # create an IVC manually so we can set the shapes.  Otherwise must set shape in the component
        # itself.
        ivc = p.model.add_subsystem('ivc', om.IndepVarComp('x', val=np.ones(xshape)))
        ivc.add_output('y', val=np.ones(yshape))
        comp = p.model.add_subsystem('comp', DotProd(derivs_method=derivs_method))
        comp.matrix_free = matrix_free
        p.model.connect('ivc.x', 'comp.x')
        p.model.connect('ivc.y', 'comp.y')

        p.setup(mode=mode, force_alloc_complex=True)

        x = np.arange(1,np.prod(xshape)+1).reshape(xshape) * 2.0
        y = np.arange(1,np.prod(yshape)+1).reshape(yshape)* 3.0
        p.set_val('ivc.x', x)
        p.set_val('ivc.y', y)
        p.final_setup()
        p.run_model()

        assert_near_equal(p.get_val('comp.z'), np.dot(x, y))
        p.check_totals(of=['comp.z'], wrt=['comp.x', 'comp.y'], method='cs', show_only_incorrect=True)
        assert_check_partials(p.check_partials(method='cs', show_only_incorrect=True))

    @parameterized.expand(itertools.product(['fwd', 'rev'], [True, False]), name_func=parameterized_name)
    def test_jax_explicit_comp2(self, mode, matrix_free):
        p = om.Problem()
        ivc = p.model.add_subsystem('ivc', om.IndepVarComp('x', val=np.ones(x_shape)))
        ivc.add_output('y', val=np.ones(y_shape))
        comp = p.model.add_subsystem('comp', DotProdMult(derivs_method='jax'))
        comp.matrix_free = matrix_free
        p.model.connect('ivc.x', 'comp.x')
        p.model.connect('ivc.y', 'comp.y')

        p.setup(mode=mode)

        x = np.arange(1,np.prod(x_shape)+1).reshape(x_shape) * 2.0
        y = np.arange(1,np.prod(y_shape)+1).reshape(y_shape)* 3.0
        p.set_val('ivc.x', x)
        p.set_val('ivc.y', y)
        p.final_setup()
        p.run_model()

        assert_near_equal(p.get_val('comp.z'), np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y * 2.5)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'],
                                           wrt=['comp.x', 'comp.y'], method='fd',
                                           show_only_incorrect=True))
        assert_check_partials(p.check_partials(show_only_incorrect=True))

    @parameterized.expand(itertools.product(['fwd', 'rev'], [True, False]), name_func=parameterized_name)
    def test_jax_explicit_comp2primal(self, mode, matrix_free):
        # this component defines its own compute_primal method
        p = om.Problem()
        ivc = p.model.add_subsystem('ivc', om.IndepVarComp('x', val=np.ones(x_shape)))
        ivc.add_output('y', val=np.ones(y_shape))
        comp = p.model.add_subsystem('comp', DotProdMultPrimal(derivs_method='jax'))
        comp.matrix_free = matrix_free
        p.model.connect('ivc.x', 'comp.x')
        p.model.connect('ivc.y', 'comp.y')

        p.setup(mode=mode)

        x = np.arange(1,np.prod(x_shape)+1).reshape(x_shape) * 2.0
        y = np.arange(1,np.prod(y_shape)+1).reshape(y_shape)* 3.0
        p.set_val('ivc.x', x)
        p.set_val('ivc.y', y)
        p.final_setup()
        p.run_model()

        assert_near_equal(p.get_val('comp.z'), np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y * 2.5)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'], wrt=['comp.x', 'comp.y'],
                                           method='fd', show_only_incorrect=True))
        assert_check_partials(p.check_partials(show_only_incorrect=True))

    @parameterized.expand(itertools.product(['fwd', 'rev'], ['fd', 'cs']), name_func=parameterized_name)
    def test_jax_explicit_comp2primal_nodecl(self, mode, derivs_method):
        # this component defines its own compute_primal method
        p = om.Problem()
        ivc = p.model.add_subsystem('ivc', om.IndepVarComp('x', val=np.ones(x_shape)))
        ivc.add_output('y', val=np.ones(y_shape))
        p.model.add_subsystem('comp', DotProdMultPrimalNoDeclPartials(derivs_method=derivs_method))
        p.model.connect('ivc.x', 'comp.x')
        p.model.connect('ivc.y', 'comp.y')

        p.setup(mode=mode, force_alloc_complex=True)

        x = np.arange(1,np.prod(x_shape)+1).reshape(x_shape) * 2.0
        y = np.arange(1,np.prod(y_shape)+1).reshape(y_shape)* 3.0
        p.set_val('ivc.x', x)
        p.set_val('ivc.y', y)
        p.final_setup()
        p.run_model()

        if derivs_method == 'fd':
            method = 'cs'
        else:
            method = 'fd'

        assert_near_equal(p.get_val('comp.z'), np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y * 2.5)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'], wrt=['comp.x', 'comp.y'],
                                           method=method, show_only_incorrect=True))
        assert_check_partials(p.check_partials(method=method, show_only_incorrect=True))

    @parameterized.expand(itertools.product(['fwd', 'rev'], [True, False]), name_func=parameterized_name)
    def test_jax_explicit_comp2primal_w_option(self, mode, matrix_free):
        # this component defines its own compute_primal method
        p = om.Problem()
        ivc = p.model.add_subsystem('ivc', om.IndepVarComp('x', val=np.ones(x_shape)))
        ivc.add_output('y', val=np.ones(y_shape))
        comp = p.model.add_subsystem('comp', DotProdMultPrimalOption(derivs_method='jax', mult=1.5))
        comp.matrix_free = matrix_free

        # deriv shape: [[(2, 4, 2, 3), (2, 4, 3, 4)], [(3, 4, 2, 3), (3, 4, 3, 4)]]
        p.model.connect('ivc.x', 'comp.x')
        p.model.connect('ivc.y', 'comp.y')

        p.setup(mode=mode)

        x = np.arange(1,np.prod(x_shape)+1).reshape(x_shape) * 2.0
        y = np.arange(1,np.prod(y_shape)+1).reshape(y_shape) * 3.0
        p.set_val('ivc.x', x)
        p.set_val('ivc.y', y)
        p.final_setup()
        p.run_model()

        assert_near_equal(p.get_val('comp.z'), np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y * 3.0)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'], wrt=['comp.x', 'comp.y'],
                                           method='fd', show_only_incorrect=True))
        assert_check_partials(p.check_partials(show_only_incorrect=True))

        comp.options['mult'] = 3.5
        p.set_val('ivc.x', x)
        p.set_val('ivc.y', y)
        p.run_model()

        assert_near_equal(p.get_val('comp.z'), np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y * 7.0)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'], wrt=['comp.x', 'comp.y'],
                                           method='fd', show_only_incorrect=True))
        assert_check_partials(p.check_partials(show_only_incorrect=True))

        comp.stat = 1./3.5
        p.set_val('ivc.x', x)
        p.set_val('ivc.y', y)
        p.run_model()

        assert_near_equal(p.get_val('comp.z'), np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'], wrt=['comp.x', 'comp.y'],
                                           method='fd', show_only_incorrect=True))
        assert_check_partials(p.check_partials(show_only_incorrect=True))

    @parameterized.expand(itertools.product(['fwd', 'rev'], [True, False]), name_func=parameterized_name)
    def test_jax_explicit_comp_with_option(self, mode, matrix_free):
        p = om.Problem()
        ivc = p.model.add_subsystem('ivc', om.IndepVarComp('x', val=np.ones(x_shape)))
        ivc.add_output('y', val=np.ones(y_shape))
        comp = p.model.add_subsystem('comp', DotProdMultOption(mult=1.7, derivs_method='jax'))
        comp.matrix_free = matrix_free

        p.model.connect('ivc.x', 'comp.x')
        p.model.connect('ivc.y', 'comp.y')

        p.setup(mode=mode)

        x = np.arange(1,np.prod(x_shape)+1).reshape(x_shape) * 2.0
        y = np.arange(1,np.prod(y_shape)+1).reshape(y_shape)* 3.0
        p.set_val('ivc.x', x)
        p.set_val('ivc.y', y)
        p.final_setup()
        p.run_model()

        assert_near_equal(p.get_val('comp.z'), np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y * 1.7)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'], wrt=['comp.x', 'comp.y'],
                                           method='fd', show_only_incorrect=True))
        assert_check_partials(p.check_partials(show_only_incorrect=True))

        p.set_val('ivc.x', x)
        p.set_val('ivc.y', y)
        comp.options['mult'] = 1.9
        p.run_model()
        assert_near_equal(p.get_val('comp.z'), np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y * 1.9)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'], wrt=['comp.x', 'comp.y'],
                                           method='fd', show_only_incorrect=True))
        assert_check_partials(p.check_partials(show_only_incorrect=True))

    @parameterized.expand(itertools.product(['fwd', 'rev'], [True, False]), name_func=parameterized_name)
    def test_jax_explicit_comp_with_discrete(self, mode, matrix_free):
        p = om.Problem()
        ivc = p.model.add_subsystem('ivc', om.IndepVarComp('x', val=np.ones(x_shape)))
        ivc.add_output('y', val=np.ones(y_shape))
        ivc.add_discrete_output('disc_out', val=3)
        comp = p.model.add_subsystem('comp', DotProductMultDiscrete(derivs_method='jax'))
        comp.matrix_free = matrix_free

        p.model.connect('ivc.x', 'comp.x')
        p.model.connect('ivc.y', 'comp.y')
        p.model.connect('ivc.disc_out', 'comp.disc_in')

        p.setup(mode=mode)

        x = np.arange(1,np.prod(x_shape)+1).reshape(x_shape) * 2.0
        y = np.arange(1,np.prod(y_shape)+1).reshape(y_shape)* 3.0
        p.set_val('ivc.x', x)
        p.set_val('ivc.y', y)
        p.final_setup()
        p.run_model()

        assert_near_equal(p.get_val('comp.z'), np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y * 3.0)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'], wrt=['comp.x', 'comp.y'],
                                           method='fd', show_only_incorrect=True))
        assert_check_partials(p.check_partials(show_only_incorrect=True))

        p.set_val('ivc.disc_out', -2)
        p.run_model()
        assert_near_equal(p.get_val('comp.z'), -np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y * 2.5)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'], wrt=['comp.x', 'comp.y'],
                                           method='fd', show_only_incorrect=True))
        assert_check_partials(p.check_partials(show_only_incorrect=True))

    @parameterized.expand(itertools.product(['fwd', 'rev'], [True, False]), name_func=parameterized_name)
    def test_jax_explicit_comp_with_discrete_primal(self, mode, matrix_free):
        p = om.Problem()
        ivc = p.model.add_subsystem('ivc', om.IndepVarComp('x', val=np.ones(x_shape)))
        ivc.add_output('y', val=np.ones(y_shape))
        ivc.add_discrete_output('disc_out', val=3)
        comp = p.model.add_subsystem('comp', DotProductMultDiscretePrimal(derivs_method='jax'))
        comp.matrix_free = matrix_free

        p.model.connect('ivc.x', 'comp.x')
        p.model.connect('ivc.y', 'comp.y')
        p.model.connect('ivc.disc_out', 'comp.disc_in')

        p.setup(mode=mode)

        x = np.arange(1,np.prod(x_shape)+1).reshape(x_shape) * 2.0
        y = np.arange(1,np.prod(y_shape)+1).reshape(y_shape)* 3.0
        p.set_val('ivc.x', x)
        p.set_val('ivc.y', y)
        p.final_setup()
        p.run_model()

        assert_near_equal(p.get_val('comp.z'), np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y * 3.0)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'], wrt=['comp.x', 'comp.y'],
                                             method='fd', show_only_incorrect=True))
        assert_check_partials(p.check_partials(show_only_incorrect=True))

        p.set_val('ivc.disc_out', -2)
        p.run_model()
        assert_near_equal(p.get_val('comp.z'), -np.dot(x, y))
        assert_near_equal(p.get_val('comp.zz'), y * 2.5)
        assert_check_totals(p.check_totals(of=['comp.z','comp.zz'], wrt=['comp.x', 'comp.y'],
                                           method='fd', show_only_incorrect=True))
        assert_check_partials(p.check_partials(show_only_incorrect=True))


if sys.version_info >= (3, 9):


    class CompRetValue(om.JaxExplicitComponent):
        def __init__(self, shape, nins=1, nouts=1, **kwargs):
            super().__init__(**kwargs)
            self.shape = shape
            self.nins = nins
            self.nouts = nouts

            self.compute_primal = getattr(self, f'compute_primal_{self.nins}_{self.nouts}')

        def setup(self):
            if self.shape == ():
                for i in range(self.nins):
                    self.add_input(f'x{i}', val=0.)
                for i in range(self.nouts):
                    self.add_output(f'y{i}', val=0.)
            else:
                for i in range(self.nins):
                    self.add_input(f'x{i}', val=jnp.zeros(self.shape))
                for i in range(self.nouts):
                    self.add_output(f'y{i}', val=jnp.zeros(self.shape))

        def setup_partials(self):
            self.declare_partials('*', '*')

        def compute_primal_1_1(self, x0):
            return x0**2

        def compute_primal_2_1(self, x0, x1):
            return x0**2 + x1**2

        def compute_primal_1_2(self, x0):
            return x0**2, x0*2

        def compute_primal_2_2(self, x0, x1):
            return x0**2, x1**2


    class CompRetTuple(om.JaxExplicitComponent):
        def __init__(self, shape, nins=1, nouts=1, **kwargs):
            super().__init__(**kwargs)
            self.shape = shape
            self.nins = nins
            self.nouts = nouts

            self.compute_primal = getattr(self, f'compute_primal_{self.nins}_{self.nouts}')

        def setup(self):
            if self.shape == ():
                for i in range(self.nins):
                    self.add_input(f'x{i}', val=0.)
                for i in range(self.nouts):
                    self.add_output(f'y{i}', val=0.)
            else:
                for i in range(self.nins):
                    self.add_input(f'x{i}', val=jnp.zeros(self.shape))
                for i in range(self.nouts):
                    self.add_output(f'y{i}', val=jnp.zeros(self.shape))

        def setup_partials(self):
            self.declare_partials('*', '*')

        def compute_primal_1_1(self, x0):
            return (x0**2,)

        def compute_primal_2_1(self, x0, x1):
            return (x0**2 + x1**2,)

        def compute_primal_1_2(self, x0):
            return x0**2, x0*2

        def compute_primal_2_2(self, x0, x1):
            return x0**2, x1**2


    class TopGrp(om.Group):
        def __init__(self, shape, ret_tuple=False, nins=1, nouts=1, matrix_free=False, **kwargs):
            super().__init__(**kwargs)
            self.shape = shape
            self.ret_tuple = ret_tuple
            self.nins = nins
            self.nouts = nouts
            self.matrix_free = matrix_free

        def setup(self):
            self.add_subsystem('ivc', om.IndepVarComp())
            if self.shape == ():
                for i in range(self.nins):
                    self.ivc.add_output(f'x{i}', 0.)

                if self.ret_tuple:
                    comp = self.add_subsystem('comp', CompRetTuple(shape=self.shape, nins=self.nins, nouts=self.nouts))
                else:
                    comp = self.add_subsystem('comp', CompRetValue(shape=self.shape, nins=self.nins, nouts=self.nouts))
            else:
                for i in range(self.nins):
                    self.ivc.add_output(f'x{i}', np.zeros(self.shape))
                if self.ret_tuple:
                    comp = self.add_subsystem('comp', CompRetTuple(shape=self.shape, nins=self.nins, nouts=self.nouts))
                else:
                    comp = self.add_subsystem('comp', CompRetValue(shape=self.shape, nins=self.nins, nouts=self.nouts))

            comp.matrix_free = self.matrix_free

            for io in range(self.nouts):
                for ii in range(self.nins):
                    if ii == io:
                        self.connect(f'ivc.x{io}', f'comp.x{ii}')



@unittest.skipIf(jax is None or sys.version_info < (3, 9), 'jax is not available or python < 3.9.')
class TestJaxShapesAndReturns(unittest.TestCase):
    @parameterized.expand(itertools.product([(), (2,), (2,3)], [(1, 1), (2, 2), (1, 2), (2, 1)],
                                            [True, False], [True, False], ['fwd', 'rev']), name_func=parameterized_name)
    def test_compute_primal_return_shapes(self, shape, sizetup, ret_tuple, matrix_free, mode):
        nins, nouts = sizetup
        prob = om.Problem()
        prob.model = TopGrp(shape=shape, ret_tuple=ret_tuple, nins=nins, nouts=nouts, matrix_free=matrix_free)
        prob.set_solver_print(level=0)

        ofs = [f'comp.y{i}' for i in range(nouts)]
        wrts = [f'ivc.x{i}' for i in range(nins)]

        prob.setup(force_alloc_complex=True, check=False, mode=mode)
        prob.final_setup()
        prob.compute_totals(of=ofs, wrt=wrts)

        assert_check_partials(prob.check_partials(method='cs', out_stream=None), atol=1e-5)
        assert_check_totals(prob.check_totals(of=ofs, wrt=wrts, method='cs', out_stream=None), atol=1e-5)

    # TODO: test with mixed np and jnp in compute

if __name__ == '__main__':
    unittest.main()



"""
Define functions and objects common to the ExplicitFuncComp and ImplicitFuncComp classes.
"""

import sys
import traceback
import re
from functools import partial
from collections.abc import Iterable

import numpy as np
try:
    import jax
    from jax import vmap
    import jax.numpy as jnp
    # linear_util moved to jax.extend in jax 0.4.17, previous location is deprecated
    try:
        from jax.extend import linear_util
    except ImportError:
        from jax import linear_util
    from jax.api_util import argnums_partial
    from jax._src.api import _jvp, _vjp
    jax.config.update("jax_enable_x64", True)  # jax by default uses 32 bit floats
except Exception:
    _, err, tb = sys.exc_info()
    if not isinstance(err, ImportError):
        traceback.print_tb(tb)
    jax = None


# regex to check for variable names.
namecheck_rgx = re.compile('[_a-zA-Z][_a-zA-Z0-9]*')

# Names that are not allowed for input or output variables (keywords for options)
_disallowed_varnames = {
    'units', 'shape', 'shape_by_conn', 'run_root_only', 'distributed', 'assembled_jac_type'
}


def _copy_with_ignore(dct, keep, ignore=()):
    """
    Copy the entries in the given dict whose keys are in keep.

    Parameters
    ----------
    dct : dict
        The dictionary to be copied.
    keep : set-like
        Set of keys for entries we want to keep.
    ignore : set or tuple
        Don't issue a warning for these non-keeper keys.

    Returns
    -------
    dict
        A new dict containing 'keep' entries.
    """
    return {k: v for k, v in dct.items() if k in keep and k not in ignore}


def _check_var_name(comp, name):
    match = namecheck_rgx.match(name)
    if match is None or match.group() != name:
        raise NameError(f"{comp.msginfo}: '{name}' is not a valid variable name.")

    if name in _disallowed_varnames:
        raise NameError(f"{comp.msginfo}: cannot use variable name '{name}' because "
                        "it's a reserved keyword.")


def jac_forward(fun, argnums, tangents):
    """
    Similar to the jax.jacfwd function but allows specification of the tangent matrix.

    This allows us to generate a compressed jacobian based on coloring.

    Parameters
    ----------
    fun : function
        The function to be differentiated.
    argnums : tuple of int or None
        Specifies which positional args are dynamic.  None means all positional args are dynamic.
    tangents : ndarray
        Array of 1.0's and 0's that is used to compute the value of the jacobian matrix.

    Returns
    -------
    function
        If there are multiple output variables, returns a function that returns rows of the
        jacobian grouped by output variable, e.g., if there were 2 output variables of size 3 and 4,
        the function would return a list with two entries. The first entry would contain the first 3
        rows of J and the second would contain the next 4 rows of J.  If there is only 1 output
        variable, the values returned are grouped by input variable.
    """
    f = linear_util.wrap_init(fun)
    if argnums is None:
        def jacfunf(*args):
            return vmap(partial(_jvp, f, args), out_axes=(None, -1))(tangents)[1]
    else:
        def jacfunf(*args):
            f_partial, dyn_args = argnums_partial(f, argnums, args)
            return vmap(partial(_jvp, f_partial, dyn_args), out_axes=(None, -1))(tangents)[1]
    return jacfunf


def jac_reverse(fun, argnums, tangents):
    """
    Similar to the jax.jacrev function but allows specification of the tangent matrix.

    This allows us to generate a compressed jacobian based on coloring.

    Parameters
    ----------
    fun : function
        The function to be differentiated.
    argnums : tuple of int or None
        Specifies which positional args are dynamic.  None means all positional args are dynamic.
    tangents : ndarray
        Array of 1.0's and 0's that is used to compute the value of the jacobian matrix.

    Returns
    -------
    function
        A function that returns rows of the jacobian grouped by function input variable, e.g., if
        there were 3 input variables of size 5 and 7 and 9, the function
        would return a list with 3 entries. The first entry would contain the first 5 columns of J,
        the second the next 7 columns of J, and the third the next 9 columns of J.  Note that for
        implicit systems, the function inputs will contain both inputs and outputs in the context
        of OpenMDAO.
    """
    f = linear_util.wrap_init(fun)
    if argnums is None:
        def jacfunr(*args):
            return vmap(_vjp(f, *args)[1])(tangents)
    else:
        def jacfunr(*args):
            f_partial, dyn_args = argnums_partial(f, argnums, args)
            return vmap(_vjp(f_partial, *dyn_args)[1])(tangents)

    return jacfunr


def jacvec_prod(fun, argnums, invals, tangent):
    """
    Similar to the jvp function but gives back a flat column.

    Note: this is significantly slower (when producing a full jacobian) than jac_forward.

    Parameters
    ----------
    fun : function
        The function to be differentiated.
    argnums : tuple of int or None
        Specifies which positional args are dynamic.  None means all positional args are dynamic.
    invals : tuple of float or ndarray
        Dynamic function input values.
    tangent : ndarray
        Array of 1.0's and 0's that is used to compute a column of the jacobian matrix.

    Returns
    -------
    function
        A function to compute the jacobian vector product.
    """
    f = linear_util.wrap_init(fun)
    if argnums is not None:
        invals = list(argnums_partial(f, argnums, invals)[1])

    # compute shaped tangents to use later
    sizes = np.array([jnp.size(a) for a in invals])
    inds = np.cumsum(sizes[:-1])
    shaped_tangents = [a.reshape(s.shape) for a, s in zip(np.split(tangent, inds, axis=0), invals)]

    if argnums is None:
        def jvfun(inps):
            return _jvp(f, inps, shaped_tangents)[1]
    else:
        def jvfun(inps):
            f_partial, dyn_args = argnums_partial(f, argnums, inps)
            return _jvp(f_partial, list(dyn_args), shaped_tangents)[1]

    return jvfun


def _get_tangents(vals, direction, coloring=None, argnums=None, trans=None):
    """
    Return a tuple of tangents values for use with vmap.

    Parameters
    ----------
    vals : list
        List of function input values.
    direction : str
        Derivative computation direction ('fwd' or 'rev').
    coloring : Coloring or None
        If not None, the Coloring object used to compute a compressed tangent array.
    argnums : list of int or None
        Indices of dynamic (differentiable) function args.
    trans : ndarray
        Translation array from jacobian indices into function arg indices.  This is needed
        because OpenMDAO expects ordering to be outputs first, then inputs, but function args
        could be in any order.

    Returns
    -------
    tuple of ndarray or ndarray
        The tangents values to be passed to vmap.
    """
    if argnums is None:
        leaves = vals
    else:
        leaves = [vals[i] for i in argnums]
    sizes = [np.size(a) for a in leaves]
    inds = np.cumsum(sizes[:-1])
    if coloring is None:
        tangent = np.eye(np.sum(sizes))
        if trans is not None:
            tangent = tangent[:, trans]
    else:
        tangent = coloring.tangent_matrix(direction, trans=trans)

    shapes = [tangent.shape[:1] + np.shape(v) for v in leaves]
    tangents = tuple([np.reshape(a, shp) for a, shp in zip(np.split(tangent, inds, axis=1),
                                                           shapes)])
    if len(leaves) == 1:
        tangents = tangents[0]

    return tangents


def _ensure_iter(val):
    """
    Turn the given value into an iterator if it is not already.

    Parameters
    ----------
    val : object
        The value to be iterated over.

    Returns
    -------
    tuple or iterable
    """
    if isinstance(val, Iterable):
        return val
    return val,
